%{

/*
nchord
program for synthesis of rational n-chords and -rhythms
Author: Klaus Wachter
e-mail: mail@klaus-wachter-1.de
Copyright: Freeware
Platform: LINUX
*/

/*
#define DEBUG
*/

#define MAX_OSCS 20

#include <fcntl.h>
#include <sys/soundcard.h>

int num;
int slash;
char *device = NULL;
FILE *audio;
int sample_rate = 44100, stereo = 0, sample_size;
int low = 0, mid = 128, high = 255; 
int osc = 0, n_oscs;
float scaling_factor;
int times[MAX_OSCS] = {2, 3}, expanded_times[MAX_OSCS], counter[MAX_OSCS],
initial[MAX_OSCS], level[MAX_OSCS], pulse_counter[MAX_OSCS];
int factor = 10;
int pulse_width = 0;
int burst_length = 0;

%}

%%

	device = realloc(device, 9);
	strcpy(device, "/dev/dsp");

	#ifdef VERIFY_AUDIO
	set_audio(stereo); fclose(audio);
	#endif

	printf("type \"?<enter>\" for help\n");

	prompt();

\x0a	prompt();

p	play();

\?	help();

^q$	printf("quit\n"); exit(0);

\"[^\"]*\"	{
	device = realloc(device, yyleng - 1);
	yytext[yyleng - 1] = 0;
	strcpy(device, yytext + 1);
	printf("device = \"%s\"\n", device);
	}

[0-9]+	{if(slash)
	{slash = 0; printf("t0 = %d\n", times[0]); osc = 1;}
	num =atoi(yytext);}

l	low = num; printf("low = %d\n", low);
m	mid = num; printf("mid = %d\n", mid);
h	high = num; printf("high = %d\n", high);

s	{sample_rate = num;
	set_audio(stereo); fclose(audio);
	printf("sample_rate = %d\n", sample_rate);}

S	{if(stereo) stereo = 0; else stereo = 1;
	printf("stereo = %d\n", stereo);}

o	if(num < MAX_OSCS) osc = num; printf("osc = %d\n", osc);

t	times[osc] = num; printf("t%d = %d\n", osc, times[osc]);

\/	slash = 1; times[0] = num;

i	{initial[osc] = num;
	printf("i%d = %d\n", osc, initial[osc]);}

f	{if(num > 0) factor = num;
	printf("factor = %d\n", factor);}

w	pulse_width = num; printf("pulse_width = %d\n", pulse_width);

b	burst_length = num; printf("burst_length = %d\n", burst_length);

\.	{
	int i;
	printf("device = \"%s\"\n", device);
	printf("num = %d, osc = %d, sample_rate = %d\n",
	num, osc, sample_rate);
	printf("factor = %d, pulse_width = %d, burst_length = %d\n",
	factor, pulse_width, burst_length);
	n_oscs = active() + 1;
	printf("n_oscs = %d\n", n_oscs);
	if(n_oscs <= 2) printf("low = %d, mid = %d, high = %d, stereo = %d\n",
	low, mid, high, stereo);
	for(i = 0; i < n_oscs; i++) {
	printf("t%d=%d ", i, times[i]);
	printf("i%d=%d ", i, initial[i]);
	}
	printf("\n");
	}

!.+$	system(yytext + 1);

#.*$	;

%%

prompt ()
{
  printf ("command> ");
}

help ()
{
  printf ("\
The program is either in mode \"command\" or \"play\".\n\
 command:\n\
  p -> mode = \"play\", start playing\n\
  ? -> display this help screen; <enter>q<enter> -> quit this program\n\
  \"<string>\" -> device = <string>\n\
  <number> -> num = <number>\n\
  s -> sample_rate = num\n\
  S -> toggle between stereo and mono\n\
  l -> low = num\n\
  m -> mid = num\n\
  h -> high = num\n\
  o -> osc = num\n\
  t -> times[osc] = num\n\
  <number1>/<number2>t -> shorthand for \"0 o <number1> t 1 o <number2> t\"\n\
  i -> initial[osc] = num\n\
  f -> factor = num\n\
  w -> pulse_width = num\n\
  b -> burst_length = num\n\
  . -> display settings\n\
  !<string><enter> -> execute the command <string>\n\
  #<string><enter> -> ignore <string> (take it as a comment)\n\
 play:\n\
  <enter> -> stop playing, mode = \"command\"\n\
");
}

unblock (fd)
     int fd;
{
  int flag;
  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag | O_NONBLOCK);
}

block (fd)
     int fd;
{
  int flag;
  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag & ~(O_NONBLOCK));
}

active ()
{
  int i;
  i = MAX_OSCS - 1;
  while (times[i] == 0 && i >= 0)
    i--;
  return (i);
}

set_audio (int mode)
{
  if ((audio = fopen (device, "w")) == NULL)
    {
      perror (device);
      exit (1);
    }
  ioctl (fileno (audio), SNDCTL_DSP_SAMPLESIZE, &sample_size);
  ioctl (fileno (audio), SNDCTL_DSP_SPEED, &sample_rate);
  ioctl (fileno (audio), SNDCTL_DSP_STEREO, &mode);
}

init_oscs ()
{
  int i;
  int expanded_initial;
  for (i = 0; i < n_oscs; i++)
    {
      expanded_times[i] = times[i] * factor;
      expanded_initial = initial[i] * factor;
      counter[i] = expanded_initial + 1;
      if (!pulse_width)
	level[i] = 0;
    }
}

rec (i)
     int i;
{
  if (counter[i] <= 0)
    return;
  if (pulse_width)
    {
      if (pulse_counter[i])
	{
	  level[i] = 1;
	  pulse_counter[i]--;
	}
      else
	level[i] = 0;
      counter[i]--;
      if (!counter[i])
	{
	  counter[i] = expanded_times[i];
	  pulse_counter[i] = pulse_width;
	}
    }
  else
    {
      counter[i]--;
      if (!counter[i])
	{
	  counter[i] = expanded_times[i];
	  level[i] = !level[i];
	}
    }
#ifdef DEBUG
  printf ("rec(%d) = %d\n", i, level[i]);
#endif
  return (level[i]);
}

nextval_8bit_mono ()
{
  rec (0);
  rec (1);
  if (!(level[0] | level[1]))
    return (low);
  else if (level[0] & level[1])
    return (high);
  else
    return (mid);
}

play_8bit_samples (n)
     int n;
{
  int i;
  if (stereo)
    {
      for (i = 0; i < n; i++)
	{
	  if (rec (0))
	    fputc (high, audio);
	  else
	    fputc (low, audio);
	  if (rec (1))
	    fputc (high, audio);
	  else
	    fputc (low, audio);
	}
    }
  else
    for (i = 0; i < n; i++)
      fputc (nextval_8bit_mono (), audio);
}

play_8bit ()
{
  char c;
  if (burst_length)
    play_8bit_samples (burst_length);
  else
    {
      printf ("type \"<enter>\" to stop playing\n");
      unblock (fileno (stdin));
      while ((c = getchar ()) == -1)
	play_8bit_samples (10000);
      block (fileno (stdin));
    }
  fclose (audio);
}

nextval_16bit ()
{
  int i;
  int added_level = 0;
  for (i = 0; i < n_oscs; i++)
    {
      if (rec (i))
	added_level += 1;
      else
	added_level -= 1;
#ifdef DEBUG
      printf ("added_level = %i\n", added_level);
#endif
    }
  return (int) (added_level * scaling_factor);
}

play_16bit_samples (n)
     int n;
{
  int i;
  int output;
  for (i = 0; i < n; i++)
    {
      output = nextval_16bit ();
#ifdef DEBUG
      printf ("output = %d\n", output);
#endif
      fwrite (&output, 2, 1, audio);
    }
}

play_16bit ()
{
  char c;
  if (burst_length)
    play_16bit_samples (burst_length);
  else
    {
      printf ("type \"<enter>\" to stop playing\n");
      unblock (fileno (stdin));
      while ((c = getchar ()) == -1)
	play_16bit_samples (10000);
      block (fileno (stdin));
    }
  fclose (audio);
}

play ()
{
  n_oscs = active () + 1;
  init_oscs ();
  if (n_oscs)
    {
      if (n_oscs <= 2)
	{
	  sample_size = 8;
	  set_audio (stereo);
	  play_8bit ();
	}
      else
	{
	  sample_size = 16;
	  set_audio (0);
	  scaling_factor = 0x7fff / n_oscs;
#ifdef DEBUG
	  printf("scaling_factor = %f\n", scaling_factor);
#endif
	  play_16bit ();
	}
    }
  else
    return;
}
